EXEC [__AddStoredProcInfo]
    /* StoredProc_Name         */ 'Vulnerabilities_CreateOrUpdateVulnerability',
    /* Internal_Indicator      */ 'Y',
    /* ReturnType_Name         */ 'void',
    /* DataTableNames_Csv      */ NULL,
    /* OutputPropertyNames_Csv */ 'UniqueId_Text,ActionTaken_Code',
    /* Description_Text        */ 'Creates or updates an analysis for a vulnerability.',
	/* Remarks_Text            */ 'ActionTaken_Code values: C - a new vulnerability was created, U - an existing vulnerability was updated, N - nothing was changed'
GO

IF OBJECT_ID('[Vulnerabilities_CreateOrUpdateVulnerability]') IS NOT NULL DROP PROCEDURE [Vulnerabilities_CreateOrUpdateVulnerability]
GO

SET QUOTED_IDENTIFIER ON
GO
SET ANSI_NULLS ON
GO

IF TYPE_ID('[VulnPackageEntry]') IS NOT NULL DROP TYPE [VulnPackageEntry]

CREATE TYPE [VulnPackageEntry] AS TABLE
(
	[PUrl_Text] VARCHAR(200) NOT NULL,
	[AffectedVersions_Text] VARCHAR(MAX)
)
GO

GRANT EXECUTE ON TYPE::[VulnPackageEntry] TO [InedoSecurityLabsUser_Role]
GO


CREATE PROCEDURE [Vulnerabilities_CreateOrUpdateVulnerability]
(
	@Ecosystem_Name VARCHAR(50),
	@Source_Name VARCHAR(20),
	@External_Id VARCHAR(100),
	@Modified_Date DATETIME,
	@Published_Date DATETIME,
	@Withdrawn_Date DATETIME,
	@Summary_Text VARCHAR(200),
	@Details_Text VARCHAR(MAX),
	@CVSSScore_Text VARCHAR(50),
	@References_Text VARCHAR(MAX),
	@CveId_Text VARCHAR(20),
	@CweNumbers_Csv VARCHAR(MAX),
	@Packages_Table dbo.VulnPackageEntry READONLY,
    @UniqueId_Text VARCHAR(20) = NULL OUT,
	@ActionTaken_Code CHAR(1) = NULL OUT
)
AS
BEGIN
    SET NOCOUNT ON

	DECLARE @Vulnerability_Id INT
	DECLARE @Current_Modified_Date DATETIME

	SELECT @Vulnerability_Id = [Vulnerability_Id],
	       @UniqueId_Text = [UniqueId_Text],
		   @Current_Modified_Date = [Modified_Date]
	  FROM [Vulnerabilities]
	 WHERE [Ecosystem_Name] = @Ecosystem_Name AND [Source_Name] = @Source_Name AND [External_Id] = @External_Id

	IF @Vulnerability_Id IS NOT NULL BEGIN

		IF @Current_Modified_Date < @Modified_Date BEGIN

			UPDATE [Vulnerabilities]
			   SET [Modified_Date] = @Modified_Date,
				   [Withdrawn_Date] = @Withdrawn_Date,
				   [Summary_Text] = @Summary_Text,
				   [Details_Text] = @Details_Text,
				   [CVSSScore_Text] = @CVSSScore_Text,
				   [References_Text] = @References_Text,
				   [CveId_Text] = @CveId_Text,
				   [CweNumbers_Csv] = @CweNumbers_Csv
			 WHERE [Vulnerability_Id] = @Vulnerability_Id

			 SET @ActionTaken_Code = 'U'

		END ELSE BEGIN

			SET @ActionTaken_Code = 'N'

		END

	END ELSE BEGIN

        -- construct a PGV-YYDDDSQ (2-digit year + "spread out" day of year + 2-digit sequence)
           -- FYI: sequence is base36-like (supports 1296 values), as were some occasional data dumps (like on 2022-05-24) that exceeded 99

		DECLARE @INTMAX INT = 2147483647

        ;WITH /* Complete and Total Overkill */

        -- Ghetto [Numbers] replacement
        Numbers AS (SELECT TOP 36 [Number] = ROW_NUMBER() OVER(ORDER BY (SELECT 1)) FROM master.dbo.spt_values),
        -- Map 0-36 to 0-Z
        CharMap AS (SELECT [Number], [Digit] = CASE WHEN [Number] <= 10 THEN CHAR([Number]+47) ELSE CHAR([Number]+54) END FROM Numbers),
        -- Map 0-1296 to 00-ZZ (like Base36 but but ordered differently)
        SeqMap AS (SELECT [Number] = ROW_NUMBER() OVER(ORDER BY ISNUMERIC(D1.[Digit] + D2.[Digit]) DESC, D1.[Number], D2.[Number]) - 1,
                          [Value] = D1.[Digit] + D2.[Digit] FROM CharMap D1, CharMap D2),
        -- Combine Existing (Date,Id) on same day with New (Date,@INTMAX)
        AH0 AS (SELECT [Vulnerability_Id], [Published_Date] FROM [Vulnerabilities]
                 WHERE CAST([Published_Date] AS DATE) = CAST(@Published_Date AS DATE)
                UNION ALL SELECT @INTMAX, @Published_Date),
        -- Add Integer Parts
        AH1 AS (SELECT [Vulnerability_Id],
                   iYY = DATEPART(YEAR, [Published_Date]) - 2000,
                   -- "Spread Out" DAYOFYEAR to "use more digits" and span 3 to 970 instead of 1 to 366
                   iDOY = CAST(ROUND(DATEPART(DAYOFYEAR, [Published_Date]) * 2.65, 0, 1) AS INT),
				   -- Get Sequence (order by date, Vuln_Id to keep a consistent ordering) 
		           iSEQ = ROW_NUMBER() OVER (PARTITION BY CAST([Published_Date] AS DATE) ORDER BY CAST([Published_Date] AS DATE), [Vulnerability_Id]) -1
                FROM AH0),
        -- Add Characters over Sequences
        AH2 AS (SELECT [Vulnerability_Id],
                   cYY = RIGHT('0' + CAST(iYY AS VARCHAR(2)), 2),
		           cDOY = RIGHT('00' + CAST(iDOY AS VARCHAR(3)), 3),
		           cSeq = SeqMap.Value
                FROM AH1 JOIN SeqMap ON AH1.iSEQ = SeqMap.Number),
        -- Munge to a Perfect ID
        UniqueIds AS (SELECT [Vulnerability_Id], [UniqueId_Text] =  'PGV-' + CYY + CDOY + cSeq  FROM AH2),
        -- Grab ID
        NextUniqueId AS (SELECT [UniqueId_Text] FROM UniqueIds WHERE [Vulnerability_Id] = @INTMAX)
		SELECT @UniqueId_Text = [UniqueId_Text] FROM NextUniqueId

        -- May be NULL if more than 1296 values, so use a GUID so we can at least insert something and hopefully fix it later
        SET @UniqueId_Text = COALESCE(@UniqueId_Text, 'PGV-' + CAST(NEWID() AS VARCHAR(36)))

		INSERT INTO [Vulnerabilities]
		(
			[UniqueId_Text],
			[Ecosystem_Name],
			[Source_Name],
			[External_Id],
			[Modified_Date],
			[Published_Date],
			[Withdrawn_Date],
			[Summary_Text],
			[Details_Text],
			[CVSSScore_Text],
			[References_Text],
			[CveId_Text],
			[CweNumbers_Csv]
		)
		VALUES
		(
			@UniqueId_Text,
			@Ecosystem_Name,
			@Source_Name,
			@External_Id,
			@Modified_Date,
			@Published_Date,
			@Withdrawn_Date,
			@Summary_Text,
			@Details_Text,
			@CVSSScore_Text,
			@References_Text,
			@CveId_Text,
			@CweNumbers_Csv
		)

		SET @Vulnerability_Id = SCOPE_IDENTITY()
		SET @ActionTaken_Code = 'C'

	END

	IF @ActionTaken_Code <> 'N' BEGIN

		UPDATE VP
		   SET VP.[AffectedVersions_Text] = P.[AffectedVersions_Text]
		  FROM [VulnerabilityPackages] VP
		       INNER JOIN @Packages_Table P
			           ON VP.[Vulnerability_Id] = @Vulnerability_Id
					  AND VP.[PUrl_Text] = P.[PUrl_Text]

		INSERT INTO [VulnerabilityPackages]
		(
			[Vulnerability_Id],
			[PUrl_Text],
			[AffectedVersions_Text]
		)
		SELECT @Vulnerability_Id,
		       P.[PUrl_Text],
			   P.[AffectedVersions_Text]
		  FROM @Packages_Table P
		 WHERE NOT EXISTS(SELECT * FROM [VulnerabilityPackages] VP WHERE VP.[Vulnerability_Id] = @Vulnerability_Id AND VP.[PUrl_Text] = P.[PUrl_Text])

		DELETE VP
		  FROM [VulnerabilityPackages] VP
		 WHERE VP.[Vulnerability_Id] = @Vulnerability_Id
		   AND NOT EXISTS(SELECT * FROM @Packages_Table P WHERE VP.[PUrl_Text] = P.[PUrl_Text])

	END

END
GO

GRANT EXECUTE ON [Vulnerabilities_CreateOrUpdateVulnerability] TO [InedoSecurityLabsUser_Role]
GO